/*
 *
 * mosplugins/ffmpeg/ffmpegImageSource.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/config.H>

#include "ffmpegImageSource.H"

#include <nucleo/utils/FileUtils.H>
#include <nucleo/image/sink/ImageSink.H>
#include <nucleo/image/encoding/Conversion.H>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <stdexcept>

// ------------------------------------------------------------------------------------------------------

extern "C" {

  void *ffmpegImageSource_factory(const URI *uri, Image::Encoding e) {
#if 0
    std::string filename = uri->isOpaque() ? uri->getOpaque() : uri->getPath() ;
    const char *extension = getExtension(filename.c_str()) ;
    if (!strcmp(extension,".mpg"))  return new ffmpegImageSource(e, *uri,"mpeg1video") ;
    if (!strcmp(extension,".mpeg")) return new ffmpegImageSource(e, *uri,"mpeg1video") ;
    throw std::runtime_error("ffmpegImageSource: unknown extension") ;
#endif
    return new ffmpegImageSource(e, *uri, "mpeg1video") ;
  }

}

// ------------------------------------------------------------------------------------------------------

static const int BufferSize = 1024*1024 ;

ffmpegImageSource::ffmpegImageSource(Image::Encoding e, const URI &uri, const char *format) {
  //_encoding = e==Image::PREFERRED ? Image::YpCbCr420 : e ;
  _encoding = Image::YpCbCr420;
  _filename = uri.isOpaque() ? uri.getOpaque() : uri.getPath() ;
  _format = format ;

  std::string query = uri.getQuery() ;
  std::string arg;
  if (URI::getQueryArg(query, "length", &arg) && arg!="")
  {
    _length = atoi(arg.c_str());
  }
  else {
    _length = -1;
  }

  if (URI::getQueryArg(query, "framerate", &arg) && arg!="")
  {
    _pace = (unsigned long)(TimeKeeper::second/atof(arg.c_str()));
  }
  else {
    _pace = (unsigned long)(TimeKeeper::second/25.0);
  }

  _state = CLOSED ;
}


bool
ffmpegImageSource::start(void) {
  if (_state!=CLOSED) return false ;

  if (!first_iformat) {
    avcodec_init();
    av_register_all() ;
    avcodec_register_all() ;
  }

  AVInputFormat *file_iformat = NULL;
  AVImageFormat *image_format = NULL;
  AVFormatParameters params, *ap = &params;
  memset(ap, 0, sizeof(*ap));
  ap->image_format = image_format;

  int err = av_open_input_file(&_ic, _filename.c_str(), file_iformat, 0, ap);

  if (err < 0) {
	  std::cerr << "ffmpegImageSource: cannot open input file: " << 
		  _filename.c_str() << " error: " << err << std::endl ;
	  return false;
  }

  err = av_find_stream_info(_ic);
  if (err < 0) {
	  std::cerr << "ffmpegImageSource: cannot find stream info"
		    << std::endl ;
	  return false;
  }

  _video_index = -1;
  for(int i = 0; i < _ic->nb_streams; i++)
  {
	  AVCodecContext *cc = &_ic->streams[i]->codec;
	  if (cc->codec_type == CODEC_TYPE_VIDEO && _video_index < 0)
	  {
		  _video_index =  i;
	  }
  }

  if (_video_index < 0)
  {
	  std::cerr << "ffmpegImageSource: no video found"
		    << std::endl ;
  }

  AVCodecContext *enc = &_ic->streams[_video_index]->codec;

  //enum CodecID id = CODEC_ID_MPEG2VIDEO ;
  // if (_format=="...") id = ... ;
  _codec = avcodec_find_decoder(enc->codec_id);
  if (!_codec) {
    std::cerr << "ffmpegImageSource: codec not found" << std::endl ;
    return false ;
  }

  // we dont send complete frames
  if(_codec->capabilities&CODEC_CAP_TRUNCATED)
	  enc->flags|= CODEC_FLAG_TRUNCATED ;
  
  if (avcodec_open(enc, _codec) < 0) {
    std::cerr << "ffmpegImageSource: could not open codec" << std::endl ;
    return false ; 
  }

  _video_st = _ic->streams[_video_index];

  _state = WAITING ;
  _tk = TimeKeeper::create(_pace, true) ;
  subscribeTo(_tk) ;
  _chrono.start() ;

  _frame = avcodec_alloc_frame(); 
  _read_len = 0;
  _is_frame_start = 1;

  dump_format(_ic, 0, _filename.c_str(), 0);
  return true ;

}

void
ffmpegImageSource::react(Observable *obs)
{
	if (obs != _tk)
	{
		return;
	}
	if (_state != WAITING)
	{
		return;
	}

	int len1, got_picture, ret;
	double pts;

	static AVPacket pkt1, *pkt = &pkt1;
	static unsigned char *ptr;
	static int64_t ipts, is_ipts;

	for (bool loop=true; loop;)
	{
		if (_read_len == 0)
		{
			ret = av_read_packet(_ic, pkt);
			if (ret < 0) {
				stop() ;
				return ;
			}
			if (pkt->stream_index != _video_index)
			{
				av_free_packet(pkt);
				continue;
			}
			ipts = pkt->pts;
			ptr = pkt->data;
			_read_len = pkt->size;
		}
		while (_read_len > 0)
		{
			if (_is_frame_start)
			{
				is_ipts = ipts;
				_is_frame_start = 0;
				ipts = AV_NOPTS_VALUE;
			}
			len1 = avcodec_decode_video(
				&_video_st->codec, _frame, &got_picture, ptr,
				_read_len);
			if (len1 < 0)
			{
				stop() ;
				return ;	
			}
			else if (len1 == 0)
			{
				//std::cerr << "ffmpegImageSource: get 0 bits" <<
				//	std::cerr;
				// ?
				loop = false;
				_read_len = 0;
				_is_frame_start = 1;
			}
			if (got_picture)
			{
				pts = 0;
				if (is_ipts != AV_NOPTS_VALUE)
					pts = (double)is_ipts *
						_ic->pts_num / _ic->pts_den;
#if 0
				if (output_picture2(is, _frame, pts) < 0)
				{
					stop() ;
					return ;
				}
#endif
				loop = false;
				_is_frame_start = 1;
			}
			ptr += len1;
			_read_len -= len1;
			if (_read_len < 0 && !got_picture)
			{
				loop = false;
				_read_len = 0;
			}
			if (loop == false)
			{
				if (_read_len > 0)
				{
					// ok
				}
				else
				{
					_read_len = 0;
				}
			}
		} // while loop
		if (_read_len == 0 && pkt)
		{
			av_free_packet(pkt);
		}
	} // for loop

	_state = READY ;
	notifyObservers() ;
}

bool
ffmpegImageSource::getNextImage(Image *img) {
  if (_state!=READY) return false ; 

  unsigned int w,h;
  w = _video_st->codec.width;
  h = _video_st->codec.height;

  unsigned char *data = 0 ;
  unsigned int datasize = 0 ;
 
  switch (_encoding) {
  case Image::L:
    img->setEncoding(Image::L) ;
    datasize = w*h ;
    data = (unsigned char *)Image::AllocMem(datasize) ;
    for(unsigned int i=0; i<h; ++i)
	 memcpy(data + i*w, _frame->data[0] + i*_frame->linesize[0], w) ;
    break ;
  case Image::YpCbCr420: 
  default: {
	 img->setEncoding(Image::YpCbCr420) ;
	 datasize = (unsigned int)(w*h*1.5) ;
	 data = (unsigned char *)Image::AllocMem(datasize) ;
	 unsigned char *yp=data, *cb=yp+w*h, *cr=cb+(int)(0.25*w*h) ;
	 unsigned int hw=w/2, hh=h/2 ;
	 for (unsigned int line=0; line<h; ++line)
	   memcpy(yp + line*w, _frame->data[0] + line*_frame->linesize[0], w) ;	 
	 for (unsigned int line=0; line<hh; ++line) {
	   memcpy(cb + line*hw, _frame->data[1] + line*_frame->linesize[1], hw) ;	 
	   memcpy(cr + line*hw, _frame->data[2] + line*_frame->linesize[2], hw) ;	 
	 }
  } break ;
  }

  img->setDims(w, h) ;
  img->setData(data, datasize, Image::FREEMEM) ;
  convertImage(img, Image::RGB) ;

  _chrono.tick() ;

  if (_length>0 && (int)(_chrono.count())==_length) {
    stop() ;
  } else {
    _state = WAITING ;
    //_tk->arm(_pace);
  }

  return true ;
}

bool
ffmpegImageSource::stop(void) {
  if (_state==CLOSED) return false ;

  _chrono.stop();

  AVCodecContext *enc = &_ic->streams[_video_index]->codec;
  avcodec_close(enc);
  av_close_input_file(_ic);
  //av_free(_ic);
  _video_st = 0;
  _video_index = -1;
  _ic = 0;
  av_free(_frame);
  _frame = 0;
  _state = CLOSED ;

  std::cerr << "ffmpegImageSink: "<< _chrono.average() << " fps"
	    << std::endl;

  unsubscribeFrom(_tk) ;
  delete _tk ;

  notifyObservers() ;
  return true ;
}
